@@ -3,19 +3,26 @@ from .commands import Redis, create_redis, create_reconnecting_redis
 from .pool import RedisPool, create_pool
 from .util import Channel
 from .errors import (
-    RedisError,
-    ProtocolError,
-    ReplyError,
-    PipelineError,
+    ConnectionClosedError,
     MultiExecError,
+    PipelineError,
+    ProtocolError,
+    RedisError,
+    ReplyError,
+    ChannelClosedError,
+    WatchVariableError,
+    PoolClosedError,
     )
 
 
-__version__ = '0.2.4'
+__version__ = '0.2.8'
 
 # make pyflakes happy
 (create_connection, RedisConnection,
  create_redis, create_reconnecting_redis, Redis,
  create_pool, RedisPool, Channel,
  RedisError, ProtocolError, ReplyError,
- PipelineError, MultiExecError)
+ PipelineError, MultiExecError, ConnectionClosedError,
+ ChannelClosedError, WatchVariableError,
+ PoolClosedError,
+ )
