@@ -1,17 +1,29 @@
 import types
 import asyncio
 import hiredis
+import socket
 from functools import partial
 from collections import deque
 
 from .util import (
     encode_command,
-    wait_ok, _NOTSET,
+    wait_ok,
+    _NOTSET,
+    _set_result,
+    _set_exception,
     coerced_keys_dict,
     Channel,
     decode,
+    async_task,
+    create_future,
+    )
+from .errors import (
+    ConnectionClosedError,
+    RedisError,
+    ProtocolError,
+    ReplyError,
+    WatchVariableError,
     )
-from .errors import RedisError, ProtocolError, ReplyError
 from .log import logger
 
 
@@ -28,7 +40,7 @@ _PUBSUB_COMMANDS = (
 
 
 @asyncio.coroutine
-def create_connection(address, *, db=None, password=None,
+def create_connection(address, *, db=None, password=None, ssl=None,
                       encoding=None, loop=None):
     """Creates redis connection.
 
@@ -38,6 +50,9 @@ def create_connection(address, *, db=None, password=None,
     * when address is a str it represents unix domain socket path.
     (no other address formats supported)
 
+    SSL argument is passed through to asyncio.create_connection.
+    By default SSL/TLS is not used.
+
     Encoding argument can be used to decode byte-replies to strings.
     By default no decoding is done.
 
@@ -49,12 +64,23 @@ def create_connection(address, *, db=None, password=None,
 
     if isinstance(address, (list, tuple)):
         host, port = address
+        logger.debug("Creating tcp connection to %r", address)
         reader, writer = yield from asyncio.open_connection(
-            host, port, loop=loop)
+            host, port, ssl=ssl, loop=loop)
+        sock = writer.transport.get_extra_info('socket')
+        if sock is not None:
+            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
+            address = sock.getpeername()
+        address = tuple(address[:2])
     else:
+        logger.debug("Creating unix connection to %r", address)
         reader, writer = yield from asyncio.open_unix_connection(
-            address, loop=loop)
-    conn = RedisConnection(reader, writer, encoding=encoding, loop=loop)
+            address, ssl=ssl, loop=loop)
+        sock = writer.transport.get_extra_info('socket')
+        if sock is not None:
+            address = sock.getpeername()
+    conn = RedisConnection(reader, writer, encoding=encoding,
+                           address=address, loop=loop)
 
     try:
         if password is not None:
@@ -71,23 +97,24 @@ def create_connection(address, *, db=None, password=None,
 class RedisConnection:
     """Redis connection."""
 
-    def __init__(self, reader, writer, *, encoding=None, loop=None):
+    def __init__(self, reader, writer, *, address, encoding=None, loop=None):
         if loop is None:
             loop = asyncio.get_event_loop()
         self._reader = reader
         self._writer = writer
+        self._address = address
         self._loop = loop
         self._waiters = deque()
         self._parser = hiredis.Reader(protocolError=ProtocolError,
                                       replyError=ReplyError)
-        self._reader_task = asyncio.Task(self._read_data(), loop=self._loop)
+        self._reader_task = async_task(self._read_data(), loop=self._loop)
         self._db = 0
         self._closing = False
         self._closed = False
-        self._close_waiter = asyncio.Future(loop=self._loop)
+        self._close_waiter = create_future(loop=self._loop)
         self._reader_task.add_done_callback(self._close_waiter.set_result)
-        self._in_transaction = False
-        self._transaction_error = None
+        self._in_transaction = None
+        self._transaction_error = None  # XXX: never used?
         self._in_pubsub = 0
         self._pubsub_channels = coerced_keys_dict()
         self._pubsub_patterns = coerced_keys_dict()
@@ -118,7 +145,7 @@ class RedisConnection:
                     # so connection must be closed
                     self._closing = True
                     self._loop.call_soon(self._do_close, exc)
-                    if self._in_transaction:
+                    if self._in_transaction is not None:
                         self._transaction_error = exc
                     return
                 else:
@@ -134,33 +161,33 @@ class RedisConnection:
     def _process_data(self, obj):
         """Processes command results."""
         waiter, encoding, cb = self._waiters.popleft()
-        if waiter.done():
-            logger.debug("Waiter future is already done %r", waiter)
-            assert waiter.cancelled(), (
-                "waiting future is in wrong state", waiter, obj)
-            return
         if isinstance(obj, RedisError):
-            waiter.set_exception(obj)
-            if self._in_transaction:
+            _set_exception(waiter, obj)
+            if self._in_transaction is not None:
                 self._transaction_error = obj
         else:
             if encoding is not None:
                 try:
                     obj = decode(obj, encoding)
                 except Exception as exc:
-                    waiter.set_exception(exc)
+                    _set_exception(waiter, exc)
                     return
-            waiter.set_result(obj)
             if cb is not None:
-                cb(obj)
+                try:
+                    obj = cb(obj)
+                except Exception as exc:
+                    _set_exception(waiter, exc)
+                    return
+            _set_result(waiter, obj)
+            if self._in_transaction is not None:
+                self._in_transaction.append((encoding, cb))
 
     def _process_pubsub(self, obj, *, _process_waiters=True):
         """Processes pubsub messages."""
         kind, *pattern, chan, data = obj
-        if _process_waiters and self._in_pubsub and self._waiters:
-            self._process_data(obj)
-
         if kind in (b'subscribe', b'unsubscribe'):
+            if _process_waiters and self._in_pubsub and self._waiters:
+                self._process_data(obj)
             if kind == b'subscribe' and chan not in self._pubsub_channels:
                 self._pubsub_channels[chan] = Channel(chan, is_pattern=False,
                                                       loop=self._loop)
@@ -170,6 +197,8 @@ class RedisConnection:
                     ch.close()
             self._in_pubsub = data
         elif kind in (b'psubscribe', b'punsubscribe'):
+            if _process_waiters and self._in_pubsub and self._waiters:
+                self._process_data(obj)
             if kind == b'psubscribe' and chan not in self._pubsub_patterns:
                 self._pubsub_patterns[chan] = Channel(chan, is_pattern=True,
                                                       loop=self._loop)
@@ -195,8 +224,8 @@ class RedisConnection:
         * ProtocolError when response can not be decoded meaning connection
           is broken.
         """
-        assert self._reader and not self._reader.at_eof(), (
-            "Connection closed or corrupted")
+        if self._reader is None or self._reader.at_eof():
+            raise ConnectionClosedError("Connection closed or corrupted")
         if command is None:
             raise TypeError("command must not be None")
         if None in set(args):
@@ -213,13 +242,15 @@ class RedisConnection:
             cb = partial(self._set_db, args=args)
         elif command in ('MULTI', b'MULTI'):
             cb = self._start_transaction
-        elif command in ('EXEC', b'EXEC', 'DISCARD', b'DISCARD'):
-            cb = self._end_transaction
+        elif command in ('EXEC', b'EXEC'):
+            cb = partial(self._end_transaction, discard=False)
+        elif command in ('DISCARD', b'DISCARD'):
+            cb = partial(self._end_transaction, discard=True)
         else:
             cb = None
         if encoding is _NOTSET:
             encoding = self._encoding
-        fut = asyncio.Future(loop=self._loop)
+        fut = create_future(loop=self._loop)
         self._writer.write(encode_command(command, *args))
         self._waiters.append((fut, encoding, cb))
         return fut
@@ -233,6 +264,8 @@ class RedisConnection:
         command = command.upper().strip()
         assert command in _PUBSUB_COMMANDS, (
             "Pub/Sub command expected", command)
+        if self._reader is None or self._reader.at_eof():
+            raise ConnectionClosedError("Connection closed or corrupted")
         if None in set(channels):
             raise TypeError("args must not contain None")
         if not len(channels):
@@ -240,7 +273,7 @@ class RedisConnection:
         cmd = encode_command(command, *channels)
         res = []
         for ch in channels:
-            fut = asyncio.Future(loop=self._loop)
+            fut = create_future(loop=self._loop)
             res.append(fut)
             self._waiters.append((fut, None, self._update_pubsub))
         self._writer.write(cmd)
@@ -267,7 +300,14 @@ class RedisConnection:
                 waiter.cancel()
             else:
                 waiter.set_exception(exc)
-        # TODO: close all subscribed channels
+        while self._pubsub_channels:
+            _, ch = self._pubsub_channels.popitem()
+            logger.debug("Closing pubsub channel %r", ch)
+            ch.close()
+        while self._pubsub_patterns:
+            _, ch = self._pubsub_patterns.popitem()
+            logger.debug("Closing pubsub pattern %r", ch)
+            ch.close()
 
     @property
     def closed(self):
@@ -280,7 +320,7 @@ class RedisConnection:
 
     @asyncio.coroutine
     def wait_closed(self):
-        yield from self._close_waiter
+        yield from asyncio.shield(self._close_waiter, loop=self._loop)
 
     @property
     def db(self):
@@ -292,6 +332,11 @@ class RedisConnection:
         """Current set codec or None."""
         return self._encoding
 
+    @property
+    def address(self):
+        """Redis server address, either host-port tuple or str."""
+        return self._address
+
     def select(self, db):
         """Change the selected database for the current connection."""
         if not isinstance(db, int):
@@ -303,29 +348,58 @@ class RedisConnection:
         return wait_ok(fut)
 
     def _set_db(self, ok, args):
-        assert ok in {b'OK', 'OK'}, ok
+        assert ok in {b'OK', 'OK'}, ("Unexpected result of SELECT", ok)
         self._db = args[0]
+        return ok
 
     def _start_transaction(self, ok):
-        assert not self._in_transaction
-        self._in_transaction = True
+        assert self._in_transaction is None, (
+            "Connection is already in transaction", self._in_transaction)
+        self._in_transaction = deque()
         self._transaction_error = None
+        return ok
 
-    def _end_transaction(self, ok):
-        assert self._in_transaction
-        self._in_transaction = False
+    def _end_transaction(self, obj, discard):
+        assert self._in_transaction is not None, (
+            "Connection is not in transaction", obj)
         self._transaction_error = None
+        recall, self._in_transaction = self._in_transaction, None
+        recall.popleft()  # ignore first (its _start_transaction)
+        if discard:
+            return obj
+        assert isinstance(obj, list) or (obj is None and not discard), (
+            "Unexpected MULTI/EXEC result", obj, recall)
+        # TODO: need to be able to re-try transaction
+        if obj is None:
+            err = WatchVariableError("WATCH variable has changed")
+            obj = [err] * len(recall)
+        assert len(obj) == len(recall), (
+            "Wrong number of result items in mutli-exec", obj, recall)
+        res = []
+        for o, (encoding, cb) in zip(obj, recall):
+            if not isinstance(o, RedisError):
+                try:
+                    if encoding:
+                        o = decode(o, encoding)
+                    if cb:
+                        o = cb(o)
+                except Exception as err:
+                    res.append(err)
+                    continue
+            res.append(o)
+        return res
 
     def _update_pubsub(self, obj):
         *head, subscriptions = obj
         self._in_pubsub, was_in_pubsub = subscriptions, self._in_pubsub
         if not was_in_pubsub:
             self._process_pubsub(obj, _process_waiters=False)
+        return obj
 
     @property
     def in_transaction(self):
         """Set to True when MULTI command was issued."""
-        return self._in_transaction
+        return self._in_transaction is not None
 
     @property
     def in_pubsub(self):
