@@ -4,14 +4,16 @@ import sys
 
 from .commands import create_redis, Redis
 from .log import logger
+from .util import async_task
+from .errors import PoolClosedError
 
 
 PY_35 = sys.version_info >= (3, 5)
 
 
 @asyncio.coroutine
-def create_pool(address, *, db=0, password=None, encoding=None,
-                minsize=10, maxsize=10, commands_factory=Redis, loop=None):
+def create_pool(address, *, db=0, password=None, ssl=None, encoding=None,
+                minsize=1, maxsize=10, commands_factory=Redis, loop=None):
     """Creates Redis Pool.
 
     By default it creates pool of commands_factory instances, but it is
@@ -26,7 +28,7 @@ def create_pool(address, *, db=0, password=None, encoding=None,
     pool = RedisPool(address, db, password, encoding,
                      minsize=minsize, maxsize=maxsize,
                      commands_factory=commands_factory,
-                     loop=loop)
+                     ssl=ssl, loop=loop)
     yield from pool._fill_free(override_min=False)
     return pool
 
@@ -36,12 +38,20 @@ class RedisPool:
     """
 
     def __init__(self, address, db=0, password=None, encoding=None,
-                 *, minsize, maxsize, commands_factory, loop=None):
+                 *, minsize, maxsize, commands_factory, ssl=None, loop=None):
+        assert isinstance(minsize, int) and minsize >= 0, (
+            "minsize must be int >= 0", minsize, type(minsize))
+        assert maxsize is not None, "Arbitrary pool size is disallowed."
+        assert isinstance(maxsize, int) and maxsize > 0, (
+            "maxsize must be int > 0", maxsize, type(maxsize))
+        assert minsize <= maxsize, (
+            "Invalid pool min/max sizes", minsize, maxsize)
         if loop is None:
             loop = asyncio.get_event_loop()
         self._address = address
         self._db = db
         self._password = password
+        self._ssl = ssl
         self._encoding = encoding
         self._minsize = minsize
         self._factory = commands_factory
@@ -50,6 +60,8 @@ class RedisPool:
         self._used = set()
         self._acquiring = 0
         self._cond = asyncio.Condition(loop=loop)
+        self._close_state = asyncio.Event(loop=loop)
+        self._close_waiter = async_task(self._do_close(), loop=loop)
 
     @property
     def minsize(self):
@@ -85,6 +97,37 @@ class RedisPool:
                 waiters.append(conn.wait_closed())
             yield from asyncio.gather(*waiters, loop=self._loop)
 
+    @asyncio.coroutine
+    def _do_close(self):
+        yield from self._close_state.wait()
+        with (yield from self._cond):
+            assert not self._acquiring, self._acquiring
+            waiters = []
+            while self._pool:
+                conn = self._pool.popleft()
+                conn.close()
+                waiters.append(conn.wait_closed())
+            for conn in self._used:
+                conn.close()
+                waiters.append(conn.wait_closed())
+            yield from asyncio.gather(*waiters, loop=self._loop)
+
+    def close(self):
+        """Close all free and in-progress connections and mark pool as closed.
+        """
+        if not self._close_state.is_set():
+            self._close_state.set()
+
+    @property
+    def closed(self):
+        """True if pool is closed."""
+        return self._close_state.is_set()
+
+    @asyncio.coroutine
+    def wait_closed(self):
+        """Wait until pool gets closed."""
+        yield from asyncio.shield(self._close_waiter, loop=self._loop)
+
     @property
     def db(self):
         """Currently selected db index."""
@@ -113,7 +156,11 @@ class RedisPool:
 
         Creates new connection if needed.
         """
+        if self.closed:
+            raise PoolClosedError("Pool is closed")
         with (yield from self._cond):
+            if self.closed:
+                raise PoolClosedError("Pool is closed")
             while True:
                 yield from self._fill_free(override_min=True)
                 if self.freesize:
@@ -136,8 +183,12 @@ class RedisPool:
         self._used.remove(conn)
         if not conn.closed:
             if conn.in_transaction:
-                logger.warning("Connection %r in transaction, closing it.",
-                               conn)
+                logger.warning(
+                    "Connection %r is in transaction, closing it.", conn)
+                conn.close()
+            elif conn.in_pubsub:
+                logger.warning(
+                    "Connection %r is in subscribe mode, closing it.", conn)
                 conn.close()
             elif conn.db == self.db:
                 if self.maxsize and self.freesize < self.maxsize:
@@ -148,7 +199,7 @@ class RedisPool:
             else:
                 conn.close()
         # FIXME: check event loop is not closed
-        asyncio.async(self._wakeup(), loop=self._loop)
+        async_task(self._wakeup(), loop=self._loop)
 
     def _drop_closed(self):
         for i in range(self.freesize):
@@ -188,6 +239,7 @@ class RedisPool:
         return create_redis(self._address,
                             db=self._db,
                             password=self._password,
+                            ssl=self._ssl,
                             encoding=self._encoding,
                             commands_factory=self._factory,
                             loop=self._loop)
